﻿var Promise = require('promise');
var diffDependencies = require('./dependencies_diff/diff-dependencies.js').diffDependencies;

function BuildDiffProcess(cfg, diffService, technologies, sourceConrolProviders, parsedDependencies) {
    
    function createBuildDiffPromiseChain(buildArguments) {
        var handlerParams = {
            buildArguments: buildArguments,
            cfg: cfg,
            finalOutput: {
                "meta": {
                    "generated": new Date().valueOf(),
                    "build": buildArguments.build,
                    "commit": buildArguments.commit,
                    "previousBuild": null,
                    "previousBranch": null,
                    "customerId": cfg.customerId,
                    "appName": buildArguments.appname,
                    "technologies": buildArguments.technology,
                    "branch": buildArguments.branch,
                    "environment": {}
                },
                brokenBuild: false, //will be set to true on parse errors, etc.
                files: [],
                dependencies: parsedDependencies || []
            },
            previousMapping: null
        };
        
        var scmInstance;
        
        if (buildArguments.workspacepath) {
            var scmType = sourceConrolProviders[buildArguments.scm];
            if (scmType) {
                scmInstance = new scmType(cfg, buildArguments.workspacePath);
            }
        }
        
        if (!buildArguments.commit && scmInstance && !scmInstance.detectCommitVersion) {
            return Promise.reject("'commit' command-line argument is required, since automatic detection by the " + buildArguments.scm + " SCM is not supported");
        }
        
        //1. Get previous mapping
        var promiseChain = diffService.getPreviousMapping(buildArguments.appname, buildArguments.branch, buildArguments.build).
        then(function getPreviousMapping(mapping) {
            if (mapping && mapping.meta && mapping.meta.build) {
                handlerParams.previousMapping = mapping;
                handlerParams.finalOutput.meta.previousBuild = mapping.meta.build;
                handlerParams.finalOutput.meta.previousBranch = mapping.meta.branch;
                diffDependencies(handlerParams.previousMapping.dependencies, handlerParams.finalOutput.dependencies);
            }
            else {
                diffDependencies([], handlerParams.finalOutput.dependencies);
            }
            return mapping;
        }); //TODO: Test what happens if this fails
        
        if (buildArguments.workspacepath) {
            
            //2. If there is a mapping, get the diffs
            if (scmInstance) {
                
                //2.1 if mapping was not found, try automatic branch detection by getting branch history from the SCM, then finding the last known mapping from that history
                promiseChain = promiseChain.then(function (mapping) {
                    if (mapping)
                        return mapping;
                    else
                        return detectBranch(handlerParams, scmInstance);
                });
                
                //2.2 if commit was not specified, automatically detect it                
                if (!buildArguments.commit) {
                    promiseChain = promiseChain.then(function () {
                        return scmInstance.detectCommitVersion(handlerParams).then(function (commit) {
                            if (commit) {
                                handlerParams.finalOutput.meta.commit = commit;
                            }
                        });
                    });
                }
                
                //2.2 Call the scm to diff from the last known state        
                promiseChain = promiseChain.then(function (commit) {
                    return scmInstance.getDiff(handlerParams);
                });
            }
            
            //3. For each technology, call its handler
            if (buildArguments.technology) {
                promiseChain = addTechnologyHandlers(promiseChain, handlerParams);
            }
        };
        
        //4. After going over all technology handlers, submit the final output to the server
        promiseChain = promiseChain.then(function () {
            if (handlerParams.finalOutput.brokenBuild) {
                return Promise.reject("Build agent encountered critical errors that prevent this build from being submitted.");
            }
            else {
                delete handlerParams.finalOutput.brokenBuild;
                return diffService.submitMapping(handlerParams.finalOutput);
            }
        });
        
        promiseChain = promiseChain.then(function () {
            console.log("Build agent finished");
            return Promise.resolve();
        }, function (err) {
            console.log("Build agent error: " + err);
            return Promise.resolve();
        });
        
        return promiseChain;
    }
    
    this.run = function run(buildArguments) {
        return createBuildDiffPromiseChain(buildArguments);
    }
    
    function addTechnologyHandlers(promiseChain, handlerParams) {
        handlerParams.buildArguments.technology.forEach(function (tech) {
            var techHandler = technologies[tech];
            promiseChain = promiseChain.then(function () {
                return techHandler(handlerParams);
            });
        });
        return promiseChain;
    }
    
    function loadBranchHistory(scmInstance, handlerParams) {
        return scmInstance.getBranchHistory(handlerParams).then(function (branchLog) {
            handlerParams.branchLog = branchLog;
            return branchLog;
        });
    }
    
    function detectBranch(handlerParams, scmInstance, buildArguments) {
        return loadBranchHistory(scmInstance, handlerParams).then(function (branchLog) {
            console.log('Branch history loaded, attempting to find it on the server');
            return diffService.findMappingForNewBranch(handlerParams, branchLog).then(function (mapping) {
                if (mapping && mapping.meta && mapping.meta.build) {
                    console.log('Previous branch is ' + mapping.meta.build + ' on ' + mapping.meta.branch);
                    handlerParams.previousMapping = mapping;
                    handlerParams.finalOutput.meta.previousBuild = mapping.meta.build;
                    handlerParams.finalOutput.meta.previousBranch = mapping.meta.branch;
                    diffDependencies(handlerParams.previousMapping.dependencies, handlerParams.finalOutput.dependencies);
                }
                else {
                    console.log('Branch is completely new to the server. All files/functions will be reported as "New"');
                    diffDependencies([], handlerParams.finalOutput.dependencies);
                }
                return mapping;
            });
        });
            
    }
}

module.exports.BuildDiffProcess = BuildDiffProcess;