﻿var Storage = require('./storage.js').Storage;
var statuses = require('../statuses.js');
var generateMethodId = require('./generate-method-id.js');

function AnonymousMethodsHandler(oldMapping, newMapping) {
	this.oldMapping = oldMapping;
	this.newMapping = newMapping;
	this.storage = new Storage();
}

AnonymousMethodsHandler.prototype.tryStore = function (method, isOldMapping) {
	if (!method.isAnonymous)
		return;
	
	var methodHash = method.hash;
	var arr;
	var lookup = this.storage.old;
	if (!isOldMapping) {
		lookup = this.storage.new;
	}
	
	arr = lookup[methodHash];
	
	if (!arr) {
		//This is the first item. Create the array.
		arr = [];
		lookup[methodHash] = arr;
	}
	
	arr.push(method);
}

AnonymousMethodsHandler.prototype.handle = function (newMapping) {
	var existingMethodsHandler = this.processMethodGroupByHash.bind(this, newMapping);
	Object.keys(this.storage.old).forEach(existingMethodsHandler);
	
	var newMethodsHandler = this.markMethodsAsAdded.bind(this);
	Object.keys(this.storage.new).forEach(newMethodsHandler);
}

AnonymousMethodsHandler.prototype.processMethodGroupByHash = function (newMapping, methodHash) {
	var oldMethodsGroupedByHash = this.storage.old[methodHash];
	var newMethodsGroupedByHash = this.storage.new[methodHash];
	
	if (!newMethodsGroupedByHash) {
		//Couldn't find anonymous method/s with the same hash in the new mapping.
		//Mark the entire methods group as deleted.
		this.markMethodsAsDeleted(oldMethodsGroupedByHash, 0, newMapping);
		return;
	}
	
	this.processSingleMethodsGroup(oldMethodsGroupedByHash, newMethodsGroupedByHash, newMapping);
}

AnonymousMethodsHandler.prototype.createLookupByPosition = function (methods) {
	var lookup = {};
	
	for (var i = 0; i < methods.length; i++) {
		var method = methods[i];
		var position = method.position.join(",");
		lookup[position] = method;
	};
	
	return lookup;
}

AnonymousMethodsHandler.prototype.processSingleMethodsGroup = function (oldMethodsArr, newMethodsArr, newMapping) {
	var oldMethodsByPosition = this.createLookupByPosition(oldMethodsArr);
	
	oldMethodsArr.sort(this.sortByLineNumber);
	newMethodsArr.sort(this.sortByLineNumber);
	
	var maxIndex = Math.min(oldMethodsArr.length, newMethodsArr.length);
	var copyOfNew = newMethodsArr.slice(0);
	for (var i = 0; i < maxIndex; i++) {
		var oldMethod = oldMethodsArr[i];
		var newMethod = copyOfNew[i];
		
		this.markMethodAsMovedOrNotChanged(oldMethod, newMethod, oldMethodsByPosition, i);
	};
	
	if (oldMethodsArr.length > copyOfNew.length) {
		//The remaining old methods are marked as deleted.
		var reminderStartIndex = copyOfNew.length;
		this.markMethodsAsDeleted(oldMethodsArr, reminderStartIndex, newMapping);
	}
}

AnonymousMethodsHandler.prototype.markMethodAsMovedOrNotChanged = function (oldMethod, newMethod, oldMethodsByPosition, index) {
	if (index === 0) {
		//Give the original method id to the first method found. 
		newMethod.id = oldMethod.id;
	}
	else {
		//All the other methods should have a unique method id.
		var salt = newMethod.hash + index;
		newMethod.id = generateMethodId(salt);
	}
	
	var positionAsString = newMethod.position.join(",");
	var hadMethodInTheSamePosition = oldMethodsByPosition[positionAsString] != null;
	if (hadMethodInTheSamePosition) {
		//Since the method has the same hash, position, it is probably the same method which hasn't changed.
		newMethod.status = statuses.METHOD_NOT_CHANGED;
	}
	else {
		newMethod.status = statuses.METHOD_MOVED;
	}
	
	this.removeMethodFromNewLookup(newMethod);

}

AnonymousMethodsHandler.prototype.sortByLineNumber = function (a, b) {
	return a.position[0] - b.position[0];
}

AnonymousMethodsHandler.prototype.markMethodsAsAdded = function (methodHash) {
	var newMethodsArrayPerHash = this.storage.new[methodHash];
	
	//method added - Generate a new ID
	for (var i = 0; i < newMethodsArrayPerHash.length; i++) {
		var salt = methodHash + i;
		var newId = generateMethodId(salt);
		newMethodsArrayPerHash[i].id = newId;
		newMethodsArrayPerHash[i].status = statuses.METHOD_ADDED;
	}
}

AnonymousMethodsHandler.prototype.markMethodsAsDeleted = function (oldMethodsArr, startIndex, newMapping) {
	for (var i = startIndex; i < oldMethodsArr.length; i++) {
		var oldMethod = oldMethodsArr[i];
		newMapping.methods.push({ id: oldMethod.id, status: statuses.METHOD_DELETED });
	};
}

AnonymousMethodsHandler.prototype.removeMethodFromNewLookup = function (method) {
	var hash = method.hash;
	var lookup = this.storage.new;
	var arr = lookup[hash];
	if (!arr) {
		return;
	}
	
	for (var i = 0; i < arr.length; i++) {
		var currentMethod = arr[i];
		if (currentMethod === method) {
			arr.splice(i, 1);
			break;
		}
	};
	
	if (arr.length === 0) {
		lookup[hash] = null;
		delete lookup[hash];
	}
}


module.exports.AnonymousMethodsHandler = AnonymousMethodsHandler;