﻿var Promise = require('promise'), 
	spawn = require('child_process').spawn, 
	path = require('path'),
	fs = require('fs'),
	statuses = require('../statuses.js'),
	generateMethodId = require('./generate-method-id.js'),
	fileDiff = require('./file-diff.js');

function javaHandler(handlerParams) {
    var opts = handlerParams.buildArguments;
    var previousMapping = handlerParams.previousMapping;
    var finalOutput = handlerParams.finalOutput;
    var diffData = handlerParams.diffData;
    
    var p = new Promise(function (resolve, reject) {
        handlerParams.cfg.javaHome = handlerParams.cfg.javaHome || process.env.JAVA_HOME;
        handlerParams.cfg.javaClassPrefix = handlerParams.cfg.javaClassPrefix || '';
        if (!handlerParams.cfg.jar) {
            console.log('jar file is missing in sealights.json. Use sl-config-cia to specify it.');
            reject(null);
            return;
        }
        var executable = path.join(handlerParams.cfg.javaHome, 'bin/java');
        console.log('Java home = ' + handlerParams.cfg.javaHome);
        console.log('JAR location = ' + handlerParams.cfg.jar);
        console.log('Java prefix = ' + handlerParams.cfg.javaClassPrefix);
        console.log('Java executable = ' + executable);
        
        var tempFilename = 'mapping-' + new Date().valueOf() + '.json';
        //find jars in path opts.workspacepath
		console.log('Searching for *.jar/*.war files in workspace path "' + opts.workspacepath + '". Recursive search: ' + opts.recursive);
		
		var options = {
			isRecursive: opts.recursive, 
			filter: isJarOrWar
		};

		searchFiles(opts.workspacepath, onSearchComplete , options);

		function onSearchComplete (err, files) {
			if (err) {
				console.log('Could not find *.jar/*.war files due to error:' + err);
				reject(err);
				return;
			}
			
			var jars = files;
			if (jars.length == 0) {
				console.log('Could not find any *.jar/*.war files in workspace, aborting');
				reject(err);
				return;
			}
			
			jars.forEach(function (fullFilename) {
				console.log('Found *.jar/*.war: ' + fullFilename)
			});
			
			try {
				var childProcess = spawn(executable, ['-jar', handlerParams.cfg.jar, tempFilename].concat(jars), { stdio: 'inherit' });
				//console.log('Spawned java pid=' + childProcess.pid);
				childProcess.on('exit', function (code, signal) {
					console.log('Java mapper existed with code=' + code + ', signal=' + signal);
					if (code || signal) {
						console.log('ERROR.');
						reject({ code: code, signal: signal });
						return;
					}
					try {
						fs.readFile(tempFilename, 'utf8', function (err, contents) {
							if (err) {
								console.log('Failed reading temp file ' + tempFilename);
								console.log('Error: ' + err);
								reject(err);
								return;
							}
							
							var newMapping;
							try {
								newMapping = JSON.parse(contents);
							}
                            catch (e) {
								console.log('Failed parsing temp file ' + tempFilename);
								console.log('Error: ' + e);
								console.log('contents: ' + contents);
								reject(e);
								return;
							}
							
							try {
								diffWithPreviousVersion(handlerParams, newMapping);
								resolve(true);
							}
                            catch (e) {
								console.log('Failed diffing between current and previous builds: ' + e);
								reject(e);
								return;
							}

						});
					} catch (e) {
						console.log('Could not read temp file ' + tempFilename + ": " + e);
						reject(e);
						return;
					}
				});
			} catch (e) {
				console.log('Failed launching java mapper:' + e);
				reject(e);
				return;
			}
		}
    });
    
    return p;
}

function isJarOrWar(filename) {
	var fileExtension = path.extname(filename).toLowerCase();
	return (fileExtension == ".jar" || fileExtension == ".war");
}


function searchFiles(dir, doneCallback, options) {
	var filterFunction = options.filter;
	var results = [];
	//console.log("[search files] Dir:" + dir);
	fs.readdir(dir, function (err, filesOrFolders) {
		//console.log("[search files] error:" + err + ", filesOrFolders:" + filesOrFolders);
		if (err) {
			return doneCallback(err);
		}

		var i = 0;

		(function processNextFileOrFolder() {
			var fileOrFolder = filesOrFolders[i++];
			if (!fileOrFolder) {
				//No more files / folders. Return the results.
				return doneCallback(null, results);
			}
			
			//Create a full path.
			fileOrFolder = path.join(dir, fileOrFolder);

			fs.stat(fileOrFolder, function (err, stat) {
				//If that's a directory, continue the search.
				if (options.isRecursive && stat && stat.isDirectory()) {
					searchFiles(fileOrFolder, function (err, res) {
						results = results.concat(res);
						processNextFileOrFolder();
					}, options);
				} else {
					//The current item is a file. Add it to the result.
					if (!filterFunction)
						results.push(fileOrFolder);
					else if (filterFunction(fileOrFolder))
						results.push(fileOrFolder);

					processNextFileOrFolder();
				}
			});
		})();
	});
};


function createFileFilter(prefix) {
    prefix = prefix || '';
    var prefixParts = prefix.split(',');
    var partCount = prefixParts.length;
    return function (filename) {
        try {
            if (partCount == 0) return true; //no filter - everything goes
            for (var i = 0; i < partCount; i++) {
                if (filename.indexOf(prefixParts[i]) === 0) {
                    return true;
                }
            }
            return false;
        } catch (e) {
            console.log(e);
            return false;
        }
    };
}

function diffWithPreviousVersion(handlerParams, newMapping) {
    var previousMapping = handlerParams.previousMapping;
    var finalOutput = handlerParams.finalOutput;
    
    finalOutput.files = finalOutput.files || [];
    fileMap = {};
    
    //map files to a hash for a quicker lookup
    finalOutput.files.forEach(function (f) {
        fileMap[f.filename] = f;
    });
    
    //map previous mapping lookup to a hash for a quicker lookup
    var previousMappingLookup = {};
    if (previousMapping && previousMapping.files) {
        previousMapping.files.forEach(function (f) {
            previousMappingLookup[f.filename] = f;
        });
    }
    
    var fileFilter = createFileFilter(handlerParams.cfg.javaClassPrefix);
	
	var totalMethods = 0;
    newMapping = newMapping || {};
    newMapping.files = newMapping.files || [];
    newMapping.files.forEach(function (file, fileIdx) {
        try {
            var filename = file.filename;
            if (fileMap[filename]) //It is possible for the same project to be in several technologies, like javascript and nodejs. These work on the same file extension and may cause duplicate mapping for the .js files.
            {
                return; //already added that file, skip it 
            }
            
            console.log('Processing [' + (fileIdx + 1) + '/' + newMapping.files.length + '] ' + filename);
            if (!fileFilter(filename)) {
                console.log('\t-->Skipped -- it does not match the specified class filter');
                //if (previousMappingLookup[filename]) {
                //    previousMappingLookup[filename].status = statuses.FILE_DELETED;
                //    previousMappingLookup[filename].methods = previousMappingLookup[filename].methods || [];
                //    previousMappingLookup[filename].methods.forEach(function (m) {m.status = statuses.METHOD_DELETED });
                //}
                //finalOutput.files.push(previousMappingLookup[filename]);
                return;
            }
            console.log('\t-->Comparing against previous version');
            if (previousMappingLookup.hasOwnProperty(filename)) {
                var prevFileMapping = previousMappingLookup[filename];
                delete previousMappingLookup[filename]; //to discover deleted files
                
                var newSig = fileDiff(prevFileMapping, file);
                fileMap[filename] = newSig;
				finalOutput.files.push(newSig);
				if (newSig && newSig.methods) {
					totalMethods += newSig.methods.length;
				}
				
            }
            else {
                file.status = statuses.FILE_ADDED;
                file.methods = file.methods || [];
                file.methods.forEach(function (m) {
					m.status = statuses.METHOD_ADDED;
					if (m.isAnonymous) {
						var salt = m.hash + fileIdx;
						m.id = generateMethodId(salt);
					}
                });
                fileMap[filename] = newSig;
				finalOutput.files.push(file);
				if (newSig && newSig.methods) {
					totalMethods += newSig.methods.length;
				}
            }
        } catch (e) {
            console.log(e);
        }
	});
	
	console.log("Processed " + totalMethods + " methods in "+newMapping.files.length +" files.");
    //console.log('-=-=-=~~~=-=-=-');
    Object.keys(previousMappingLookup).forEach(function (deletedFilename) {
        //Handle deletion of [deletedFilename]        
        var deletedFile = previousMappingLookup[deletedFilename];
        if (deletedFile.status = statuses.FILE_DELETED)
            return; //Skip an already-deleted file
        console.log('DELETED: ' + deletedFile.filename);
        deletedFile.status = statuses.FILE_DELETED;
        deletedFile.methods = deletedFile.methods || [];
        deletedFile.methods.forEach(function (m) {
            m.status = statuses.METHOD_DELETED;
        });
        finalOutput.files.push(deletedFile);
    });
}

module.exports = javaHandler;